from ..model import Slab
from ..polymer import PolymerBrush
from .polymeri import PolymerBrushInteractor
from ..mono import FreeInterface, FreeLayer
from .monoi import FreeInterfaceInteractor, FreeLayerInteractor
from .layer import NoInteractor, SlabInteractor


# ======================== LayerInteractor factory ====================
# Associate layers with layer interactors through function
#     interactor(profile,layer)
# New layer interactors can be registered with
#     make_interactor[layer_class] = interactor_class
# =====================================================================
class InteractorFactory:
    """
    Given a layer, find the associated interactor.
    """

    def __init__(self):
        self._registry = {}

    def __setitem__(self, layer_class, interactor_class):
        self._registry[layer_class] = interactor_class

    def __call__(self, profile, layer):
        constructor = self._registry.get(layer.__class__, NoInteractor)
        return constructor(profile, layer)


interactor = InteractorFactory()
interactor[Slab] = SlabInteractor
interactor[PolymerBrush] = PolymerBrushInteractor
interactor[FreeInterface] = FreeInterfaceInteractor
interactor[FreeLayer] = FreeLayerInteractor



import wx

from .model_view import ModelView


def make_frame(experiment):
    frame = wx.Frame(None, -1, "profile interactor")
    panel = ModelView(frame)
    panel.quit_on_error()
    panel.set_job(experiment)
    frame.panel = panel
    return frame


def demo(experiment):
    app = wx.PySimpleApp(0)
    frame = make_frame(experiment)
    frame.Show(True)
    app.MainLoop()


def example_experiment():
    import numpy as np
    from ..names import Experiment, NeutronProbe, Material, silicon, air

    T = np.linspace(0, 3, 100)
    probe = NeutronProbe(L=4.75, dL=4.75 * 0.02, T=T, dT=T * 0.01)
    chrome = Material("Cr")
    gold = Material("Au")
    sample = silicon(0, 5) | chrome(80, 5) | gold(50, 5) | air
    return Experiment(probe=probe, sample=sample)


if __name__ == "__main__":
    demo(example_experiment())



import math


class BinIter:
    def __init__(self, n, edges):
        self.n = n
        self.edges = edges
        self.forward = edges[0] < edges[n]

        if self.forward:
            self.bin = 0
            self.lo = edges[0]
            self.hi = edges[1]
        else:
            self.bin = n - 1
            self.lo = edges[n]
            self.hi = edges[n - 1]

        self.atend = n < 1

    def increment(self):
        if self.atend:
            raise IndexError("moving beyond final bin")
        self.lo = self.hi
        if self.forward:
            self.bin += 1
            self.atend = self.bin >= self.n
            if not self.atend:
                self.hi = self.edges[self.bin + 1]
        else:
            self.atend = self.bin == 0
            if not self.atend:
                self.bin -= 1
                self.hi = self.edges[self.bin]
        return self


def rebin_counts_portion(Nold, vold, Iold, Nnew, vnew, Inew, ND_portion):
    # Note: inspired by rebin from OpenGenie, but using counts per bin
    # rather than rates.

    # Does not work in place
    if Iold is Inew:
        raise ValueError("does not work in place")

    # Traverse both sets of bin edges; if there is an overlap, add the portion
    # of the overlapping old bin to the new bin.
    _from = BinIter(Nold, vold)
    _to = BinIter(Nnew, vnew)
    while not _from.atend and not _to.atend:
        # std::cout << "from " << from.bin << ": [" << from.lo << ", " << from.hi << "]\n";
        # std::cout << "to " << to.bin << ": [" << to.lo << ", " << to.hi << "]\n";
        if _to.hi <= _from.lo:
            _to.increment()  # new must catch up to old
        elif _from.hi <= _to.lo:
            _from.increment()  # old must catch up to new
        else:
            overlap = min(_from.hi, _to.hi) - max(_from.lo, _to.lo)
            portion = overlap / (_from.hi - _from.lo)
            Inew[_to.bin] += Iold[_from.bin] * portion * ND_portion
            if _to.hi > _from.hi:
                _from.increment()
            else:
                _to.increment()


def rebin_counts_old(Nold, xold, Iold, Nnew, xnew, Inew):
    # Note: inspired by rebin from OpenGenie, but using counts per bin
    # rather than rates.

    # Clear the new bins
    for i in range(Nnew):
        Inew[i] = 0

    rebin_counts_portion(Nold, xold, Iold, Nnew, xnew, Inew, 1.0)


def rebin_counts(xold, Iold, xnew, Inew):
    # Note: inspired by rebin from OpenGenie, but using counts per bin
    # rather than rates.
    Nold = len(Iold)
    Nnew = len(Inew)
    # Clear the new bins
    for i in range(Nnew):
        Inew[i] = 0

    rebin_counts_portion(Nold, xold, Iold, Nnew, xnew, Inew, 1.0)


def rebin_intensity(Nold, xold, Iold, dIold, Nnew, xnew, Inew, dInew):
    # Note: inspired by rebin from OpenGenie, but using counts per bin rather than rates.

    # Clear the new bins
    for i in range(Nnew):
        dInew[i] = Inew[i] = 0

    # Traverse both sets of bin edges; if there is an overlap, add the portion
    # of the overlapping old bin to the new bin.
    _from = BinIter(Nold, xold)
    _to = BinIter(Nnew, xnew)
    while not _from.atend and not _to.atend:
        # std::cout << "from " << from.bin << ": [" << from.lo << ", " << from.hi << "]\n";
        # std::cout << "to " << to.bin << ": [" << to.lo << ", " << to.hi << "]\n";
        if _to.hi <= _from.lo:
            _to.increment()  # new must catch up to old
        elif _from.hi <= _to.lo:
            _from.increment()  # old must catch up to new
        else:
            overlap = min(_from.hi, _to.hi) - max(_from.lo, _to.lo)
            portion = overlap / (_from.hi - _from.lo)

            Inew[_to.bin] += (Iold[_from.bin]) * portion
            # add in quadrature
            dInew[_to.bin] += (dIold[_from.bin] * portion) ** 2
            if _to.hi > _from.hi:
                _from.increment()
            else:
                _to.increment()

    # Convert variance to standard deviation.
    for i in range(Nnew):
        dInew[i] = math.sqrt(dInew[i])


def rebin_counts_2D(xold, yold, Iold, xnew, ynew, Inew):
    Nxold = len(xold) - 1
    Nyold = len(yold) - 1
    Nxnew = len(xnew) - 1
    Nynew = len(ynew) - 1

    # Clear the new bins
    for i in range(Nxnew):
        for j in range(Nynew):
            Inew[i][j] = 0

    # Traverse both sets of bin edges; if there is an overlap, add the portion
    # of the overlapping old bin to the new bin.  Scale this by the portion
    # of the overlap in y.
    _from = BinIter(Nxold, xold)
    _to = BinIter(Nxnew, xnew)

    while not _from.atend and not _to.atend:
        if _to.hi <= _from.lo:
            _to.increment()  # new must catch up to old
        elif _from.hi <= _to.lo:
            _from.increment()  # old must catch up to new
        else:
            overlap = min(_from.hi, _to.hi) - max(_from.lo, _to.lo)
            portion = overlap / (_from.hi - _from.lo)
            rebin_counts_portion(Nyold, yold, Iold[_from.bin], Nynew, ynew, Inew[_to.bin], portion)
            if _to.hi > _from.hi:
                _from.increment()
            else:
                _to.increment()



# This program is public domain
# Author: Paul Kienzle
r"""
Magnetic modeling for 1-D reflectometry.

** Deprecated ** use magnetism to set magnetism on a nuclear layer
rather than creating a magnetic layer with a nuclear layer stack
underneath.

Magnetic properties are tied to the structural description of the
but only loosely.

There may be dead regions near the interfaces of magnetic materials.

Magnetic behaviour may be varying in complex ways within and
across structural boundaries.  For example, the ma
Indeed, the pattern may continue
across spacer layers, going to zero in the magnetically dead
region and returning to its long range variation on entry to
the next magnetic layer.  Magnetic multilayers may exhibit complex
magnetism throughout the repeated section while the structural
components are fixed.

The scattering behaviour is dependent upon net field strength relative to
polarization direction.   This arises from three underlying quantities:
the strength of the individual dipole moments in the layer, the degree
of alignment of these moments, and the net direction of the alignment.
The strength of the dipole moment depends on the details of the electronic
structure, so unlike the nuclear scattering potential, it cannot be readily
determined from material composition.  Similarly, net magnetization
depends on the details of the magnetic domains within the material, and
cannot readily be determined from first principles.  The interaction
potential of the net magnetic moment depends on the alignment of the field
with respect to the beam, with a net scattering length density of
:math:`\rho_M \cos(\theta_M)`.  Clearly the scattering measurement will
not be able to distinguish between a reduced net magnetic strength
:math:`\rho_M` and a change in orientation :math:`\theta_M` for an
individual measurement, as should be apparent from the correlated
uncertainty plot produced when both parameters are fit.

Magnetism support is split into two parts: describing the layers
and anchoring them to the structure.
"""

from __future__ import print_function

import numpy as np
from numpy import inf
from bumps.parameter import Parameter, to_dict
from bumps.mono import monospline

from .model import Layer, Stack
from .reflectivity import BASE_GUIDE_ANGLE as DEFAULT_THETA_M


class MagneticLayer(Layer):
    @property
    def ismagnetic(self):
        return True

    def __init__(
        self, stack=None, dead_below=0, dead_above=0, interface_below=None, interface_above=None, name="magnetic"
    ):
        self.stack = Stack(stack)
        self.dead_below = Parameter.default(dead_below, limits=(0, inf), name=name + " dead below")
        self.dead_above = Parameter.default(dead_above, limits=(0, inf), name=name + " dead above")
        self.interface_below = interface_below
        self.interface_above = interface_above
        self.name = name

    def parameters(self):
        return {
            "stack": self.stack.parameters(),
            "dead_below": self.dead_below,
            "dead_above": self.dead_above,
            "interface_below": self.interface_below,
            "interface_above": self.interface_above,
        }

    def to_dict(self):
        return to_dict(
            {
                "type": type(self).__name__,
                "name": self.name,
                "stack": self.stack,
                "dead_below": self.dead_below,
                "dead_above": self.dead_above,
                "interface_below": self.interface_below,
                "interface_above": self.interface_above,
            }
        )

    @property
    def thickness(self):
        """Thickness of the magnetic region"""
        return self.stack.thickness

    def render_stack(self, probe, slabs):
        """
        Render the nuclear sld structure.

        If either the interface below or the interface above is left
        unspecified, the corresponding nuclear interface is used.

        Returns the anchor point in the nuclear structure and interface
        widths at either end of the magnetic slab.
        """
        anchor = slabs.thickness() + self.dead_below.value

        s_below = self.interface_below.value if self.interface_below else slabs.surface_sigma
        self.stack.render(probe, slabs)
        s_above = self.interface_above.value if self.interface_above else slabs.surface_sigma
        return anchor, (s_below, s_above)

    def penalty(self):
        return self.stack.penalty()

    @property
    def thicknessM(self):
        return self.stack.thickness - (self.dead_below.value + self.dead_above.value)


class MagneticSlab(MagneticLayer):
    """
    Region of constant magnetism.
    """

    def __init__(self, stack, rhoM=0, thetaM=DEFAULT_THETA_M, name="magnetic", **kw):
        MagneticLayer.__init__(self, stack=stack, name=name, **kw)
        self.rhoM = Parameter.default(rhoM, name=name + " SLD")
        self.thetaM = Parameter.default(thetaM, limits=(0, 360), name=name + " angle")

    def parameters(self):
        parameters = MagneticLayer.parameters(self)
        parameters.update(rhoM=self.rhoM, thetaM=self.thetaM)
        return parameters

    def to_dict(self):
        ret = MagneticLayer.to_dict(self)
        ret["rhoM"] = to_dict(self.rhoM)
        ret["thetaM"] = to_dict(self.thetaM)
        return ret

    def render(self, probe, slabs):
        anchor, sigma = self.render_stack(probe, slabs)
        slabs.add_magnetism(
            anchor=anchor, w=[self.thicknessM], rhoM=[self.rhoM.value], thetaM=[self.thetaM.value], sigma=sigma
        )

    def __str__(self):
        return "magnetic(%g)" % self.rhoM.value

    def __repr__(self):
        return "Magnetic(rhoM=%g, thetaM=%g)" % (self.rhoM.value, self.thetaM.value)


class MagneticStack(MagneticLayer):
    """
    Magnetic slabs within a magnetic layer.
    """

    def __init__(self, stack, weight=(), rhoM=(), thetaM=(DEFAULT_THETA_M,), interfaceM=(0), name="mag. stack", **kw):
        if (
            len(thetaM) != 1
            and len(thetaM) != len(weight)
            and len(rhoM) != 1
            and len(rhoM) != len(weight)
            and len(interfaceM) != 1
            and len(interfaceM) != len(weight) - 1
        ):
            raise ValueError("Must have one rhoM, thetaM and intefaceM for each layer")
        if interfaceM != [0]:
            raise NotImplementedError("Doesn't support magnetic roughness")

        MagneticLayer.__init__(self, stack=stack, name=name, **kw)
        self.weight = [Parameter.default(v, name=name + " weight[%d]" % i) for i, v in enumerate(weight)]
        self.rhoM = [Parameter.default(v, name=name + " rhoM[%d]" % i) for i, v in enumerate(rhoM)]
        self.thetaM = [Parameter.default(v, name=name + " angle[%d]" % i) for i, v in enumerate(thetaM)]
        self.interfaceM = [Parameter.default(v, name=name + " interface[%d]" % i) for i, v in enumerate(interfaceM)]

    def parameters(self):
        parameters = MagneticLayer.parameters(self)
        parameters.update(rhoM=self.rhoM, thetaM=self.thetaM, interfaceM=self.interfaceM, weight=self.weight)
        return parameters

    def to_dict(self):
        ret = MagneticLayer.to_dict(self)
        ret["rhoM"] = to_dict(self.rhoM)
        ret["thetaM"] = to_dict(self.thetaM)
        ret["interfaceM"] = to_dict(self.interfaceM)
        ret["weight"] = to_dict(self.weight)
        return ret

    def render(self, probe, slabs):
        anchor, sigma = self.render_stack(probe, slabs)
        w = np.array([p.value for p in self.weight])
        w *= self.thicknessM.value / np.sum(w)
        rhoM = [p.value for p in self.rhoM]
        thetaM = [p.value for p in self.thetaM]
        sigmaM = [p.value for p in self.interfaceM]
        if len(rhoM) == 1:
            rhoM = [rhoM[0]] * len(w)
        if len(thetaM) == 1:
            thetaM = [thetaM[0]] * len(w)
        if len(sigmaM) == 1:
            sigmaM = [sigmaM[0]] * (len(w) - 1)

        slabs.add_magnetism(anchor=anchor, w=w, rhoM=rhoM, thetaM=thetaM, sigma=sigma)

    def __str__(self):
        return "MagneticStack(%d)" % (len(self.rhoM))

    def __repr__(self):
        return "MagneticStack"


class MagneticTwist(MagneticLayer):
    """
    Linear change in magnetism throughout layer.
    """

    magnetic = True

    def __init__(self, stack, rhoM=(0, 0), thetaM=(DEFAULT_THETA_M, DEFAULT_THETA_M), name="twist", **kw):
        MagneticLayer.__init__(self, stack=stack, name=name, **kw)
        self.rhoM = [Parameter.default(v, name=name + " rhoM[%d]" % i) for i, v in enumerate(rhoM)]
        self.thetaM = [Parameter.default(v, name=name + " angle[%d]" % i) for i, v in enumerate(thetaM)]

    def parameters(self):
        parameters = MagneticLayer.parameters(self)
        parameters.update(rhoM=self.rhoM, thetaM=self.thetaM)
        return parameters

    def to_dict(self):
        ret = MagneticLayer.to_dict(self)
        ret["rhoM"] = to_dict(self.rhoM)
        ret["thetaM"] = to_dict(self.thetaM)
        return ret

    def render(self, probe, slabs):
        anchor, sigma = self.render_stack(probe, slabs)
        w, z = slabs.microslabs(self.thicknessM)
        rhoM = np.linspace(self.rhoM[0].value, self.rhoM[1].value, len(z))
        thetaM = np.linspace(self.thetaM[0].value, self.thetaM[1].value, len(z))
        slabs.add_magnetism(anchor=anchor, w=w, rhoM=rhoM, thetaM=thetaM, sigma=sigma)

    def __str__(self):
        return "twist(%g->%g)" % (self.rhoM[0].value, self.rhoM[1].value)

    def __repr__(self):
        return "MagneticTwist"


class FreeMagnetic(MagneticLayer):
    """
    Linear change in magnetism throughout layer.
    """

    magnetic = True

    def __init__(self, stack, z=(), rhoM=(), thetaM=(), name="freemag", **kw):
        MagneticLayer.__init__(self, stack=stack, name=name, **kw)

        def parvec(vector, name, limits):
            return [Parameter.default(p, name=name + "[%d]" % i, limits=limits) for i, p in enumerate(vector)]

        self.rhoM, self.thetaM, self.z = [
            parvec(v, name + " " + part, limits)
            for v, part, limits in zip((rhoM, thetaM, z), ("rhoM", "angle", "z"), ((0, inf), (0, 360), (0, 1)))
        ]
        if len(self.z) != len(self.rhoM):
            raise ValueError("must have number of intervals dz one less than rhoM")
        if len(self.thetaM) > 0 and len(self.rhoM) != len(self.thetaM):
            raise ValueError("must have one thetaM for each rhoM")

    def parameters(self):
        parameters = MagneticLayer.parameters(self)
        parameters.update(rhoM=self.rhoM, thetaM=self.thetaM, z=self.z)
        return parameters

    def to_dict(self):
        ret = MagneticLayer.to_dict(self)
        ret["rhoM"] = to_dict(self.rhoM)
        ret["thetaM"] = to_dict(self.thetaM)
        ret["z"] = to_dict(self.z)
        return ret

    def profile(self, Pz):
        thickness = self.thickness.value
        mbelow, tbelow = 0, (self.thetaM[0].value if self.thetaM else DEFAULT_THETA_M)
        mabove, tabove = 0, (self.thetaM[-1].value if self.thetaM else DEFAULT_THETA_M)
        z = np.sort([0] + [p.value for p in self.z] + [1]) * thickness

        rhoM = np.hstack((mbelow, [p.value for p in self.rhoM], mabove))
        PrhoM = monospline(z, rhoM, Pz)

        if np.any(np.isnan(PrhoM)):
            print("in mono")
            print("z %s" % str(z))
            print("p %s" % str([p.value for p in self.z]))

        if len(self.thetaM) > 1:
            thetaM = np.hstack((tbelow, [p.value for p in self.thetaM], tabove))
            PthetaM = monospline(z, thetaM, Pz)
        elif len(self.thetaM) == 1:
            PthetaM = np.full_like(PrhoM, self.thetaM.value)
        else:
            PthetaM = np.full_like(PrhoM, DEFAULT_THETA_M)
        return PrhoM, PthetaM

    def render(self, probe, slabs):
        anchor, sigma = self.render_stack(probe, slabs)
        Pw, Pz = slabs.microslabs(self.thicknessM)
        rhoM, thetaM = self.profile(Pz)
        slabs.add_magnetism(anchor=anchor, w=Pw, rhoM=rhoM, thetaM=thetaM, sigma=sigma)

    def __str__(self):
        return "freemag(%d)" % (len(self.rhoM))

    def __repr__(self):
        return "FreeMagnetic"



from __future__ import division
from math import log
import numpy
from .interactor import BaseInteractor
from .layer import MaterialInteractor
from .config import pick_radius
from .config import profile_color
from .config import vf_scale
from .util import setpar


# -----------------------------------------------------------------------
class PolymerBrushInteractor(BaseInteractor):
    """
    Interactor for tethered polymer model.
    """

    def __init__(self, profile, layer):
        super(PolymerBrushInteractor, self).__init__(profile)
        self.polymer = MaterialInteractor(profile, layer.polymer, range=(0, 0.3))
        self.solvent = MaterialInteractor(profile, layer.solvent, range=(0.7, 1))
        self.layer = layer
        ax = profile.axes

        style = dict(
            linestyle=":",
            transform=profile.xcoords,
            zorder=0,
            color=profile_color,
            visible=True,
        )
        self.hprofile = ax.plot([], [], **style)[0]

        style = dict(
            linestyle="-",
            transform=profile.xcoords,
            zorder=4,
            pickradius=pick_radius,
            color=profile_color,
            linewidth=1,
            visible=True,
        )
        self.hphi = ax.plot([], [], **style)[0]
        self.hbase = ax.plot([], [], **style)[0]
        self.hlength = ax.plot([], [], **style)[0]

        style = dict(
            marker="s",
            transform=profile.xcoords,
            zorder=5,
            pickradius=pick_radius,
            color=profile_color,
            alpha=0.5,
            markersize=5,
            visible=True,
        )
        self.hpower = ax.plot([], [], **style)[0]
        self.hsigma = ax.plot([], [], **style)[0]
        # print "xcoords",id(profile.xcoords),"trans",id(self.hbase.get_transform())

        self.markers = [self.hphi, self.hbase, self.hlength, self.hpower, self.hsigma]
        self.parameters = [layer.base_vf, layer.base, layer.length, layer.power, layer.sigma]
        self.connect_markers(self.markers)

    def clear_markers(self):
        super(PolymerBrushInteractor, self).clear_markers()
        self.polymer.clear_markers()
        self.solvent.clear_markers()
        self.hprofile.remove()

    def update_markers(self):
        """
        Draw the widgets in their new positions.
        """
        self.polymer.update_markers()
        self.solvent.update_markers()

        n = self.profile.layer_num
        left = self.profile.boundary[n]
        layer = self.layer

        # TODO: want to use smae step size for profile as we do for
        # smooth profiles  Don't know how to access that value nicely.
        z = numpy.arange(0, layer.thickness.value, 0.1)
        vf = layer.profile(z)
        self.hprofile.set_data(z + left, vf * vf_scale)

        # at z midway between l1 and l2
        #     y = phi * 0.75**p
        phi = layer.base_vf.value * vf_scale / 100
        phi_power = phi * 0.75**layer.power.value
        L0 = left + layer.base.value
        L1 = L0 + layer.length.value
        Lpower = (L0 + L1) / 2
        Lsigma = L1 + layer.sigma.value

        # print "L0,L1,Lpower,Lsigma",L0,L1,Lpower,Lsigma,phi,phi_power
        self.hphi.set_data((left, L0), (phi, phi))
        self.hbase.set_data((L0, L0), (0, phi))
        self.hlength.set_data((L1, L1), (0, phi))
        self.hpower.set_data((Lpower,), (phi_power,))
        self.hsigma.set_data((Lsigma,), (0,))

    def drag(self, ev):
        """
        Update the model with the new widget position.
        """
        table = dict(zip(self.markers, self.parameters))
        par = table[ev.artist]
        n = self.profile.layer_num
        left = self.profile.boundary[n]
        right = self.profile.boundary[n + 1]
        # TODO: Keeping individual interactors within the layer does not
        # necessarily keep the whole interactor in the layer.
        z = max(left, min(right, ev.xdata))
        if ev.artist == self.hphi:
            # print ev.ydata, self.profile.xcoords.inverted().transform([(ev.x,ev.y)])[0][1]
            setpar(par, ev.ydata * 100 / vf_scale)
            # print "phi",par.value,ev.ydata*100/vf_scale
        elif ev.artist == self.hbase:
            offset = left
            setpar(par, z - offset)
        elif ev.artist == self.hlength:
            offset = left + self.layer.base.value
            setpar(par, z - offset)
        elif ev.artist == self.hsigma:
            offset = left + self.layer.base.value + self.layer.length.value
            setpar(par, z - offset)
        elif ev.artist == self.hpower:
            phi = self.layer.base_vf.value * vf_scale / 100
            phi_power = log(ev.ydata / phi) / log(0.75) if ev.ydata > 0 else 100
            setpar(par, phi_power)

    def save(self, ev):
        """
        Save the current state of the model represented by the widget.
        """
        table = dict(zip(self.markers, self.parameters))
        par = table[ev.artist]
        self._save_par = par
        self._save_val = par.value

    def restore(self, ev):
        """
        Restore the widget and model to the saved state.
        """
        self._save_par.value = self._save_val



import asyncio
from aiohttp import web
from dataclasses import dataclass
import json
from pathlib import Path
import socket
from typing import Union, Dict, List, Optional

import numpy as np
from bumps.webview.server import webserver
from bumps.webview.server.webserver import (
    start_app,
    sio,
    main,
    create_server_task,
    display_inline_jupyter,
    open_tab_link,
)

# Register the refl1d model loader
import refl1d.fitplugin
import bumps.cli

bumps.cli.install_plugin(refl1d.fitplugin)

from . import api  # use side-effects to register refl1d functions
from .profile_plot import plot_sld_profile_plotly

webserver.CDN_TEMPLATE = "https://cdn.jsdelivr.net/npm/refl1d-webview-client@{client_version}/dist/{client_version}"
webserver.CLIENT_PATH = Path(__file__).parent.parent / "client"


@dataclass
class Refl1DOptions(webserver.BumpsOptions):
    serializer: webserver.SERIALIZERS = "dataclass"
    headless: bool = True


webserver.OPTIONS_CLASS = Refl1DOptions
webserver.APPLICATION_NAME = "refl1d"

if __name__ == "__main__":
    main()



"""
Layer interactor.
"""

from __future__ import division
from bumps.parameter import Parameter
from refl1d.material import SLD
from .interactor import BaseInteractor
from .config import pick_radius
from .config import rho_color, rhoI_color, disabled_color
from .util import setpar


class MaterialInteractor(BaseInteractor):
    def __init__(self, profile, material, range=(0, 1)):  # @ReservedAssignment
        super(MaterialInteractor, self).__init__(profile)
        self.material = material
        self.range = range

        if isinstance(material, SLD):
            enabled = True
            self._rho = material.rho
            self._rhoI = material.irho
            self._rho_scale = 1
            self._rhoI_scale = 1
        elif hasattr(material, "density") and isinstance(material.density, Parameter):
            enabled = True
            rho, rhoI = material.sld(profile.experiment.probe)
            self._rho = material.density
            self._rhoI = material.density
            self._rho_scale = 1 / (rho / material.density.value)
            self._rhoI_scale = 1 / (rhoI / material.density.value)
        else:
            self._rho = self._rhoI = None
            enabled = False

        style = dict(
            linestyle="--",
            linewidth=2,
            pickradius=pick_radius,
        )
        ax = profile.axes
        rho = ax.plot(
            [], [], label=material.name + " rho", color=rho_color if enabled else disabled_color, zorder=6, **style
        )[0]
        rhoI = ax.plot(
            [], [], label=material.name + " rhoI", color=rhoI_color if enabled else disabled_color, zorder=5, **style
        )[0]

        self.markers = [rho, rhoI]
        if enabled:
            self.connect_markers(self.markers)

    def update_markers(self):
        z = self.profile.boundary[1:-1]
        n = self.profile.layer_num
        # if n is None: return
        if n == 0:
            left, right = -20, 0
        elif n >= len(z):
            left, right = z[-1], z[-1] + 20
        else:
            delta = z[n] - z[n - 1]
            left, right = self.range[0] * delta + z[n - 1], self.range[1] * delta + z[n - 1]

        rho, rhoI = self.material.sld(self.profile.experiment.probe)
        self.markers[0].set_data((left, right), (rho, rho))
        self.markers[1].set_data((left, right), (rhoI, rhoI))

    def save(self, ev):
        if self._rho is not None:
            self._rho_save = self._rho.value
        if self._rhoI is not None:
            self._rhoI_save = self._rhoI.value

    def restore(self, ev):
        if self._rho is not None:
            self._rho.value = self._rho_save
        if self._rhoI is not None:
            self._rhoI.value = self._rhoI_save

    def drag(self, ev):
        if ev.artist == self.markers[0]:
            setpar(self._rho, ev.ydata * self._rho_scale)
        else:
            setpar(self._rhoI, ev.ydata * self._rhoI_scale)


class SlabInteractor(BaseInteractor):
    def __init__(self, profile, layer):
        super(SlabInteractor, self).__init__(profile)
        self.material = MaterialInteractor(profile, layer.material)

    def update_markers(self):
        self.material.update_markers()

    def clear_markers(self):
        self.material.clear_markers()


# -------------------------------------------------------------------
class NoInteractor(BaseInteractor):
    """
    Null Interactor for undefined layers.
    """

    def __init__(self, profile, layer):
        super(NoInteractor, self).__init__(profile)

    # TODO: turn layer.parameters() into sliders



# This program is in the public domain
# Author: Paul Kienzle
r"""
Reflectometry instrument definitions.

An instrument definition contains all the information necessary to compute
the resolution for a measurement.  See :mod:`resolution` for details.

This module is intended to help define new instrument loaders

Scanning Reflectometers
=======================

:mod:`refl1d.instrument` (this module) defines two instrument types:
:class:`Monochromatic` and :class:`Pulsed`. These represent
generic scanning and time of flight instruments, respectively.

To perform a simulation or load a data set, a measurement geometry must
be defined.  In the following example, we set up the geometry for a
pretend instrument SP:2. The complete geometry needs to include information
to calculate wavelength resolution (wavelength and wavelength dispersion)
as well as angular resolution (slit distances and openings, and perhaps
sample size and sample warp).  In this case, we are using a scanning
monochromatic instrument with slits of 0.1 mm below 0.5\ |deg| and
opening slits  above 0.5\ |deg| starting at 0.2 mm.  The monochromatic
instrument assumes a fixed $\Delta \theta / \theta$ while opening.

    >>> from refl1d.names import *
    >>> geometry = Monochromatic(instrument="SP:2", radiation="neutron",
    ...    wavelength=5.0042, dLoL=0.009, d_s1=230+1856, d_s2=230,
    ...    Tlo=0.5, slits_at_Tlo=0.2, slits_below=0.1)

This instrument can be used to  a data file, or generate a
measurement probe for use in modeling or to read in a previously
measured data set or generate a probe for simulation:

    >>> from numpy import linspace, loadtxt
    >>> datafile = sample_data('10ndt001.refl')
    >>> Q, R, dR = loadtxt(datafile).T
    >>> probe = geometry.probe(Q=Q, data=(R, dR))
    >>> simulation = geometry.probe(T=linspace(0, 5, 51))

All instrument parameters can be specified when constructing the probe,
replacing the defaults that are associated with the instrument.  For
example, to include sample broadening effects in the resolution:

    >>> probe2 = geometry.probe(Q=Q, data=(R, dR), sample_broadening=0.1,
    ...    name="probe2")

For magnetic systems a polarized beam probe is needed::

    >>> magnetic_probe = geometry.magnetic_probe(T=np.linspace(0, 5, 100))

The string representation of the geometry prints a multi-line
description of the default instrument configuration:

    >>> print(geometry)
    == Instrument SP:2 ==
    radiation = neutron at 5.0042 Angstrom with 0.9% resolution
    slit distances = 2086 mm and 230 mm
    fixed region below 0.5 and above 90 degrees
    slit openings at Tlo are 0.2 mm
    sample width = 1e+10 mm
    sample broadening = 0 degrees

Predefined Instruments
======================

Specific instruments can be defined for each facility.  This saves the users
having to remember details of the instrument geometry.

For example, the above SP:2 instrument could be defined as follows:

    >>> class SP2(Monochromatic):
    ...    instrument = "SP:2"
    ...    radiation = "neutron"
    ...    wavelength = 5.0042   # Angstroms
    ...    dLoL = 0.009          # FWHM
    ...    d_s1 = 230.0 + 1856.0 # mm
    ...    d_s2 = 230.0          # mm
    ...    def load(self, filename, **kw):
    ...        Q, R, dR = loadtxt(datafile).T
    ...        probe = self.probe(Q=Q, data=(R, dR), **kw)
    ...        return probe

This definition can then be used to define the measurement geometry.  We
have added a load method which knows about the facility file format (in
this case, three column ASCII data Q, R, dR) so that we can load a datafile
in a couple of lines of code:

    >>> geometry = SP2(Tlo=0.5, slits_at_Tlo=0.2, slits_below=0.1)
    >>> probe3 = geometry.load(datafile)

The defaults() method prints the static components of the geometry:

    >>> print(SP2.defaults())
    == Instrument class SP:2 ==
    radiation = neutron at 5.0042 Angstrom with 0.9% resolution
    slit distances = 2086 mm and 230 mm

GUI Usage
=========

Graphical user interfaces follow different usage patterns from scripts.
Here the emphasis will be on selecting a data set to process, displaying
its default metadata and allowing the user to override it.

File loading should follow the pattern established in reflectometry
reduction, with an extension registry and a fallback scheme whereby
files can be checked in a predefined order.  If the file cannot be
loaded, then the next loader is tried.  This should be extended with
the concept of a magic signature such as those used by graphics and
sound file applications: read the first block and run it through
the signature check before trying to load it.  For unrecognized
extensions, all loaders can be tried.

The file loader should return an instrument instance with metadata
initialized from the file header.  This metadata can be displayed
to the user along with a plot of the data and the resolution.  When
metadata values are changed, the resolution can be recomputed and the
display updated.  When the data set is accepted, the final resolution
calculation can be performed.
"""

from __future__ import division, print_function

# TODO: the resolution calculator should not be responsible for loading
# the data; maybe do it as a mixin?

import numpy as np
# from numpy import pi, inf, sqrt, log, degrees, radians, cos, sin, tan

from .resolution import QL2T
from .resolution import bins, binwidths, binedges
from .resolution import slit_widths, divergence
from .probe import make_probe, PolarizedNeutronProbe
from .reflectivity import BASE_GUIDE_ANGLE


class Monochromatic(object):
    """
    Instrument representation for scanning reflectometers.

    :Parameters:
        *instrument* : string
            name of the instrument
        *radiation* : string | xray or neutron
            source radiation type
        *d_s1*, *d_s2* : float | mm
            distance from sample to pre-sample slits 1 and 2; post-sample
            slits are ignored
        *wavelength* : float | |Ang|
            wavelength of the instrument
        *dLoL* : float
            constant relative wavelength dispersion; wavelength range and
            dispersion together determine the bins
        *slits* : float OR (float, float) | mm
            fixed slits
        *slits_at_Tlo* : float OR (float, float) | mm
            slit 1 and slit 2 openings at Tlo; this can be a scalar if both
            slits are open by the same amount, otherwise it is a pair (s1, s2).
        *slits_at_Qlo* : float OR (float, float) | mm
            equivalent to slits_at_Tlo, for instruments that are controlled by
            Q rather than theta
        *Tlo*, *Thi* : float | |deg|
            range of opening slits, or inf if slits are fixed.
        *Qlo*, *Qhi* : float | |1/Ang|
            range of opening slits when instrument is controlled by Q.
        *slits_below*, *slits_above* : float OR (float, float) | mm
            slit 1 and slit 2 openings below Tlo and above Thi; again, these
            can be scalar if slit 1 and slit 2 are the same, otherwise they
            are each a pair (s1, s2).  Below and above default to the values of
            the slits at Tlo and Thi respectively.
        *sample_width* : float | mm
            width of sample; at low angle with tiny samples, stray neutrons
            miss the sample and are not reflected onto the detector, so the
            sample itself acts as a slit, therefore the width of the sample
            may be needed to compute the resolution correctly
        *sample_broadening* : float | |deg| FWHM
            amount of angular divergence (+) or focusing (-) introduced by
            the sample; this is caused by sample warp, and may be read off
            of the rocking curve by subtracting (s1+s2)/2/(d_s1-d_s2) from
            the FWHM width of the rocking curve
    """

    instrument = "monochromatic"
    radiation = "unknown"
    # Required attributes
    wavelength = None
    dLoL = None
    d_s1 = None
    d_s2 = None
    # Optional attributes
    Tlo = 90  # Use 90 for fixed slits; this is effectively inf
    Thi = 90
    fixed_slits = None
    slits_at_Tlo = None  # Slit openings at Tlo, and default for slits_below
    slits_below = None  # Slit openings below Tlo, or fixed slits if Tlo=90
    slits_above = None
    sample_width = 1e10  # Large but finite value
    sample_broadening = 0

    def __init__(self, **kw):
        self._translate_Q_to_theta(kw)
        for k, v in kw.items():
            if hasattr(self, k):
                setattr(self, k, v)
            else:
                raise TypeError("unexpected keyword argument '%s'" % k)

    def probe(self, **kw):
        """
        Return a probe for use in simulation.

        :Parameters:
            *Q* : [float] | |Ang|
                Q values to be measured.
            *T* : [float] | |deg|
                Angles to be measured.

        Additional keyword parameters

        :Returns:
            *probe* : Probe
                Measurement probe with complete resolution information.  The
                probe will not have any data.

        If both *Q* and *T* are specified then *Q* takes precedents.

        You can override instrument parameters using key=value.  In
        particular, settings for *slits_at_Tlo*, *Tlo*, *Thi*,
        *slits_below*, and *slits_above* are used to define the
        angular divergence.
        """
        self._translate_Q_to_theta(kw)
        T, dT, L, dL = self.resolution(**kw)
        sample_broadening = kw.get("sample_broadening", self.sample_broadening)
        kw.update(T=T, dT=dT - sample_broadening, L=L, dL=dL)
        kw.setdefault("radiation", self.radiation)
        return make_probe(**kw)

    def magnetic_probe(self, Aguide=BASE_GUIDE_ANGLE, shared_beam=True, H=0, **kw):
        """
        Simulate a polarized measurement probe.

        Returns a probe with Q, angle, wavelength and the associated
        uncertainties, but not any data.

        Guide field angle *Aguide* can be specified, as well as keyword
        arguments for the geometry of the probe cross sections such as
        *slits_at_Tlo*, *Tlo*, *Thi*, *slits_below*, and *slits_above*
        to define the angular divergence.
        """
        base_name = kw.pop("name", "")
        probes = [self.probe(name=base_name + xs, **kw) for xs in ("--", "-+", "+-", "++")]
        probe = PolarizedNeutronProbe(probes, Aguide=Aguide, H=H)
        if shared_beam:
            probe.shared_beam()  # Share the beam parameters by default
        return probe

    def resolution(self, **kw):
        """
        Calculate resolution at each angle.

        :Return:
            *T*, *dT* : [float] | |deg|
                Angles and angular divergence.
            *L*, *dL* : [float] | |Ang|
                Wavelengths and wavelength dispersion.
        """
        self._translate_Q_to_theta(kw)
        if "T" not in kw:
            raise TypeError("resolution requires slits and either T or Q")

        L = kw.get("L", kw.get("wavelength", self.wavelength))
        if "dL" in kw:
            dL = kw["dL"]
        else:
            dL = kw.get("dLoL", self.dLoL) * L

        if L is None or dL is None:
            raise TypeError("Need wavelength and wavelength dispersion to compute resolution")

        T = kw["T"]
        if "dT" in kw:
            dT = kw["dT"]
        else:
            if "slits" not in kw:
                kw["slits"] = self.calc_slits(**kw)
            dT = self.calc_dT(**kw)

        return T, dT, L, dL

    def calc_slits(self, **kw):
        """
        Determines slit openings from measurement pattern.

        If slits are fixed simply return the same slits for every angle,
        otherwise use an opening range [Tlo, Thi] and the value of the
        slits at the start of the opening to define the slits.  Slits
        below Tlo and above Thi can be specified separately.

        *T* OR *Q*       incident angle or Q
        *Tlo*, *Thi*     angle range over which slits are opening
        *slits_at_Tlo*   openings at the start of the range, or fixed opening
        *slits_below*, *slits_above*   openings below and above the range

        Use fixed_slits is available, otherwise use opening slits.
        """
        self._translate_Q_to_theta(kw)
        if "T" not in kw:
            raise TypeError("calc_slits requires angle T=... or Q=...")
        T = kw["T"]
        Tlo = kw.get("Tlo", self.Tlo)
        Thi = kw.get("Thi", self.Thi)
        fixed_slits = kw.get("fixed_slits", self.fixed_slits)
        if fixed_slits is not None:
            slits_at_Tlo = slits_below = slits_above = fixed_slits
            Tlo = 90
        else:
            slits_at_Tlo = kw.get("slits_at_Tlo", self.slits_at_Tlo)
            slits_below = kw.get("slits_below", self.slits_below)
            slits_above = kw.get("slits_above", self.slits_above)

        # Otherwise we are using opening slits
        if Tlo is None or slits_at_Tlo is None:
            raise TypeError("Resolution calculation requires Tlo and slits_at_Tlo")
        slits = slit_widths(
            T=T, slits_at_Tlo=slits_at_Tlo, Tlo=Tlo, Thi=Thi, slits_below=slits_below, slits_above=slits_above
        )
        return slits

    def calc_dT(self, **kw):
        """
        Compute the angular divergence for given slits and angles

        :Parameters:
            *T* OR *Q* : [float] | |deg| OR |1/Ang|
                measurement angles
            *slits* : float OR (float, float) | mm
                total slit opening from edge to edge, not beam center to edge
            *d_s1*, *d_s2* : float | mm
                distance from sample to slit 1 and slit 2
            *sample_width* : float | mm
                size of sample
            *sample_broadening* : float | |deg| FWHM
                resolution changes from sample warp

        :Returns:
            *dT* : [float] | |deg| FWHM
                angular divergence

        *sample_broadening* can be estimated from W, the full width at half
        maximum of a rocking curve measured in degrees:

            sample_broadening = W - degrees( 0.5*(s1+s2) / (d1-d2))

        """
        self._translate_Q_to_theta(kw)
        if "T" not in kw or "slits" not in kw:
            raise TypeError("calc_dT requires slits and either T or Q")
        slits = kw["slits"]
        T = kw["T"]
        d_s1 = kw.get("d_s1", self.d_s1)
        d_s2 = kw.get("d_s2", self.d_s2)
        if d_s1 is None or d_s2 is None:
            raise TypeError("Need slit distances d_s1, d_s2 to compute resolution")
        sample_width = kw.get("sample_width", self.sample_width)
        sample_broadening = kw.get("sample_broadening", self.sample_broadening)
        dT = divergence(
            T=T, slits=slits, distance=(d_s1, d_s2), sample_width=sample_width, sample_broadening=sample_broadening
        )

        return dT

    def _translate_Q_to_theta(self, kw):
        """
        Rewrite keyword arguments with Q values translated to theta values.
        """
        # Grab wavelength first so we can translate Qlo/Qhi to Tlo/Thi no
        # matter what order the keywords appear.
        wavelength = kw.get("wavelength", self.wavelength)
        if "Q" in kw:
            kw["T"] = QL2T(kw.pop("Q"), wavelength)
        if "Qlo" in kw:
            kw["Tlo"] = QL2T(kw.pop("Qlo"), wavelength)
        if "Qhi" in kw:
            kw["Thi"] = QL2T(kw.pop("Qhi"), wavelength)
        if "slits_at_Qlo" in kw:
            kw["slits_at_Tlo"] = kw.pop("slits_at_Qlo")

    def __str__(self):
        msg = """\
== Instrument %(name)s ==
radiation = %(radiation)s at %(L)g Angstrom with %(dLpercent)g%% resolution
slit distances = %(d_s1)g mm and %(d_s2)g mm
fixed region below %(Tlo)g and above %(Thi)g degrees
slit openings at Tlo are %(slits_at_Tlo)s mm
sample width = %(sample_width)g mm
sample broadening = %(sample_broadening)g degrees\
""" % dict(
            name=self.instrument,
            L=self.wavelength,
            dLpercent=self.dLoL * 100,
            d_s1=self.d_s1,
            d_s2=self.d_s2,
            sample_width=self.sample_width,
            sample_broadening=self.sample_broadening,
            Tlo=self.Tlo,
            Thi=self.Thi,
            slits_at_Tlo=str(self.slits_at_Tlo),
            radiation=self.radiation,
        )
        return msg

    @classmethod
    def defaults(cls):
        """
        Return default instrument properties as a printable string.
        """
        msg = """\
== Instrument class %(name)s ==
radiation = %(radiation)s at %(L)s Angstrom with %(dL)s resolution
slit distances = %(d_s1)s mm and %(d_s2)s mm"""
        L, d_s1, d_s2 = ["%g" % v if v is not None else "unknown" for v in (cls.wavelength, cls.d_s1, cls.d_s2)]
        dL = "%g%%" % (cls.dLoL * 100) if cls.dLoL else "unknown"
        return msg % dict(name=cls.instrument, radiation=cls.radiation, L=L, dL=dL, d_s1=d_s1, d_s2=d_s2)


class Pulsed(object):
    """
    Instrument representation for pulsed reflectometers.

    :Parameters:
        *instrument* : string
            name of the instrument
        *radiation* : string | xray, neutron
            source radiation type
        *TOF_range* : (float, float)
            usabe range of times for TOF data
        *T* : float | |deg|
            sample angle
        *d_s1*, *d_s2* : float | mm
            distance from sample to pre-sample slits 1 and 2; post-sample
            slits are ignored
        *wavelength* : (float, float) | |Ang|
            wavelength range for the measurement
        *dLoL* : float
            constant relative wavelength dispersion; wavelength range and
            dispersion together determine the bins
        *slits* : float OR (float, float) | mm
            fixed slits
        *slits_at_Tlo* : float OR (float, float) | mm
            slit 1 and slit 2 openings at Tlo; this can be a scalar if both
            slits are open by the same amount, otherwise it is a pair (s1, s2).
        *Tlo*, *Thi* : float | |deg|
            range of opening slits, or inf if slits are fixed.
        *slits_below*, *slits_above* : float OR (float, float) | mm
            slit 1 and slit 2 openings below Tlo and above Thi; again, these
            can be scalar if slit 1 and slit 2 are the same, otherwise they
            are each a pair (s1, s2).  Below and above default to the values of
            the slits at Tlo and Thi respectively.
        *sample_width* : float | mm
            width of sample; at low angle with tiny samples, stray neutrons
            miss the sample and are not reflected onto the detector, so the
            sample itself acts as a slit, therefore the width of the sample
            may be needed to compute the resolution correctly
        *sample_broadening* : float | |deg| FWHM
            amount of angular divergence (+) or focusing (-) introduced by
            the sample; this is caused by sample warp, and may be read off
            of the rocking curve by subtracting 0.5*(s1+s2)/(d_s1-d_s2) from
            the FWHM width of the rocking curve
    """

    instrument = "pulsed"
    radiation = "neutron"  # unless someone knows how to do TOF Xray...
    # Required attributes
    d_s1 = None
    d_s2 = None
    slits = None
    T = None
    wavelength = None
    dLoL = None  # usually 0.02 for 2% FWHM
    # Optional attributes
    TOF_range = (0, np.inf)
    Tlo = 90  # Use 90 for fixed slits; this is effectively inf
    Thi = 90
    fixed_slits = None
    slits_at_Tlo = None  # Slit openings at Tlo, and default for slits_below
    slits_below = None  # Slit openings below Tlo, or fixed slits if Tlo=90
    slits_above = None
    sample_width = 1e10  # Large but finite value
    sample_broadening = 0

    def __init__(self, **kw):
        for k, v in kw.items():
            if not hasattr(self, k):
                raise TypeError("unexpected keyword argument '%s'" % k)
            setattr(self, k, v)

    def probe(self, **kw):
        """
        Simulate a measurement probe.

        Returns a probe with Q, angle, wavelength and the associated
        uncertainties, but not any data.

        You can override instrument parameters using key=value.
        In particular, slit settings *slits* and *T* define
        the angular divergence and *dLoL* defines the wavelength
        resolution.
        """
        low, high = kw.get("wavelength", self.wavelength)
        dLoL = kw.get("dLoL", self.dLoL)
        T = kw.pop("T", self.T)
        L = bins(low, high, dLoL)
        dL = binwidths(L)
        T, dT, L, dL = self.resolution(L=L, dL=dL, T=T, **kw)
        sample_broadening = kw.get("sample_broadening", self.sample_broadening)
        return make_probe(T=T, dT=dT - sample_broadening, L=L, dL=dL, radiation=self.radiation, **kw)

    def magnetic_probe(self, Aguide=BASE_GUIDE_ANGLE, shared_beam=True, **kw):
        """
        Simulate a polarized measurement probe.

        Returns a probe with Q, angle, wavelength and the associated
        uncertainties, but not any data.

        Guide field angle *Aguide* can be specified, as well as keyword
        arguments for the geometry of the probe cross sections such as
        slit settings *slits* and *T* to define the angular divergence
        and *dLoL* to define the wavelength resolution.
        """
        probes = [self.probe(**kw) for _ in range(4)]
        probe = PolarizedNeutronProbe(probes, Aguide=Aguide)
        if shared_beam:
            probe.shared_beam()  # Share the beam parameters by default
        return probe

    def simulate(self, sample, uncertainty=1, **kw):
        """
        Simulate a run with a particular sample.

        :Parameters:
            *sample* : Stack
                Reflectometry model
            *T* : [float] | |deg|
                List of angles to be measured, such as [0.15, 0.4, 1, 2].
            *slits* : [float] or [(float, float)] | mm
                Slit settings for each angle.
            *uncertainty* = 1 : float or [float] | %
                Incident intensity is set so that the median dR/R is equal
                to *uncertainty*, where R is the idealized reflectivity
                of the sample.
            *dLoL* = 0.02: float
                Wavelength resolution
            *normalize* = True : boolean
                Whether to normalize the intensities
            *theta_offset* = 0 : float | |deg|
                Sample alignment error
            *background* = 0 : float
                Background counts per incident neutron (background is
                assumed to be independent of measurement geometry).
            *back_reflectivity* = False : boolean
                Whether beam travels through incident medium
                or through substrate.
            *back_absorption* = 1 : float
                Absorption factor for beam traveling through substrate.
                Only needed for back reflectivity measurements.
        """
        from numpy.random import poisson as pois
        from .rebin import rebin
        from .experiment import Experiment
        from .probe import ProbeSet

        T = kw.pop("T", self.T)
        slits = kw.pop("slits", self.slits)
        if slits is None:
            raise TypeError("Need slit openings for simulation")

        dLoL = kw.pop("dLoL", self.dLoL)
        normalize = kw.pop("normalize", True)
        theta_offset = kw.pop("theta_offset", 0)
        background = kw.pop("background", 0)
        back_reflectivity = kw.pop("back_reflectivity", False)
        back_absorption = kw.pop("back_absorption", 1)
        uncertainty *= 0.01

        # Compute reflectivity with resolution and added noise
        probes = []
        for Ti, Si in zip(T, slits):
            probe = self.probe(T=Ti, slits=Si, dLoL=dLoL)
            probe.back_reflectivity = back_reflectivity
            probe.theta_offset.value = theta_offset
            probe.back_absorption.value = back_absorption
            M = Experiment(probe=probe, sample=sample)
            _, Rth = M.reflectivity()
            # Note: probe.L is reversed because L is sorted by increasing
            # Q in probe.
            I = rebin(binedges(self.feather[0]), self.feather[1], binedges(probe.L[::-1]))[::-1]
            Ci = np.median(1.0 / (uncertainty**2 * I * Rth))
            Igoal = Ci * I
            Ibeam = pois(Igoal) + 1.0

            Irefl = pois(Igoal * Rth) + 1.0
            if background > 0:
                Irefl += pois(Igoal * background) + 1.0
                Iback = pois(Igoal * background) + 1.0
            else:
                Iback = 0
            # Set intensity/background _after_ calculating the theory function
            # since we don't want the theory function altered by them.
            probe.background.value = background
            # Correct for the feather.  This has to be done otherwise we
            # won't see the correct reflectivity.  Even if corrected for
            # the feather, though, we haven't necessarily corrected for
            # the overall number of counts in the measurement.
            # Z = X/Y
            # var Z = ( (var X / X)**2 + (var Y / Y)**2 ) * Z**2
            #       = (1/X + 1/Y) * (X/Y)**2
            #       = (Y + X) * X/Y**3
            #    dZ = sqrt( (Y+X)*X/Y**3) = sqrt((Y+X)*(X/Y))/Y
            R = (Irefl - Iback) / Ibeam
            dR = np.sqrt((Irefl + Iback + Ibeam) * (Irefl / Ibeam)) / Ibeam
            if np.isnan(dR).any() or (dR == 0).any() or (R <= 0).any():
                print("Ibeam %s" % Ibeam)
                print("Irefl %s" % Irefl)
                print("Iback %s" % Iback)
                print("R %s" % R)
                print("dR %s" % dR)
                raise RuntimeError("Should not be able to get here!!")
            # dR[Irefl==0] == 1./Ibeam[Irefl==0]
            # print("median %s"%np.median(dR/R))

            if not normalize:
                # Ci = 1./max(R)
                R, dR = R * Ci, dR * Ci
                probe.background.value *= Ci
                probe.intensity.value = Ci

            probe.R = R
            probe.dR = dR
            probes.append(probe)

        return Experiment(sample=sample, probe=ProbeSet(probes))

    def resolution(self, L, dL, **kw):
        """
        Return the resolution of the measurement.  Needs *T*, *L*, *dL*
        specified as keywords.
        """
        T = kw.pop("T", self.T)
        if "slits" not in kw:
            kw["slits"] = self.calc_slits(T=T, **kw)
        dT = self.calc_dT(T, **kw)

        # Compute the FWHM angular divergence in radians
        # Return the resolution
        return T, dT, L, dL

    def calc_slits(self, **kw):
        """
        Determines slit openings from measurement pattern.

        If slits are fixed simply return the same slits for every angle,
        otherwise use an opening range [Tlo, Thi] and the value of the
        slits at the start of the opening to define the slits.  Slits
        below Tlo and above Thi can be specified separately.

        *T*              incident angle
        *Tlo*, *Thi*     angle range over which slits are opening
        *slits_at_Tlo*   openings at the start of the range, or fixed opening
        *slits_below*, *slits_above*   openings below and above the range

        Use fixed_slits is available, otherwise use opening slits.
        """
        if "T" not in kw:
            raise TypeError("calc_slits requires angle T=...")
        T = kw["T"]
        Tlo = kw.get("Tlo", self.Tlo)
        Thi = kw.get("Thi", self.Thi)
        fixed_slits = kw.get("fixed_slits", self.fixed_slits)
        if fixed_slits is not None:
            slits_at_Tlo = slits_below = slits_above = fixed_slits
            Tlo = 90
        else:
            slits_at_Tlo = kw.get("slits_at_Tlo", self.slits_at_Tlo)
            slits_below = kw.get("slits_below", self.slits_below)
            slits_above = kw.get("slits_above", self.slits_above)

        # Otherwise we are using opening slits
        if Tlo is None or slits_at_Tlo is None:
            raise TypeError("Resolution calculation requires Tlo and slits_at_Tlo")
        slits = slit_widths(
            T=T, slits_at_Tlo=slits_at_Tlo, Tlo=Tlo, Thi=Thi, slits_below=slits_below, slits_above=slits_above
        )
        return slits

    def calc_dT(self, T, slits, **kw):
        d_s1 = kw.get("d_s1", self.d_s1)
        d_s2 = kw.get("d_s2", self.d_s2)
        sample_width = kw.get("sample_width", self.sample_width)
        sample_broadening = kw.get("sample_broadening", self.sample_broadening)
        dT = divergence(
            T=T, slits=slits, distance=(d_s1, d_s2), sample_width=sample_width, sample_broadening=sample_broadening
        )

        return dT

    def __str__(self):
        msg = """\
== Instrument %(name)s ==
radiation = %(radiation)s in %(L_min)g to %(L_max)g Angstrom with %(dLpercent)g%% resolution
slit distances = %(d_s1)g mm and %(d_s2)g mm
slit openings = %(slits)s mm
sample width = %(sample_width)g mm
sample broadening = %(sample_broadening)g degrees FWHM
""" % dict(
            name=self.instrument,
            L_min=self.wavelength[0],
            L_max=self.wavelength[1],
            dLpercent=self.dLoL * 100,
            d_s1=self.d_s1,
            d_s2=self.d_s2,
            slits=str(self.slits),
            sample_width=self.sample_width,
            sample_broadening=self.sample_broadening,
            radiation=self.radiation,
        )
        return msg

    @classmethod
    def defaults(cls):
        """
        Return default instrument properties as a printable string.
        """
        msg = """\
== Instrument class %(name)s ==
radiation = %(radiation)s in %(L_min)g to %(L_max)g Angstrom with %(dLpercent)g%% resolution
slit distances = %(d_s1)g mm and %(d_s2)g mm
""" % dict(
            name=cls.instrument,
            L_min=cls.wavelength[0],
            L_max=cls.wavelength[1],
            dLpercent=cls.dLoL * 100,
            d_s1=cls.d_s1,
            d_s2=cls.d_s2,
            radiation=cls.radiation,
        )
        return msg


_ = '''
class GenericMonochromatic(Monochromatic):
    def load(self, filename, **kw):
        """
        Load the data, returning the associated probe.  This probe will
        contain Q, angle, wavelength, measured reflectivity and the
        associated uncertainties.

        You can override instrument parameters using key=value.  In
        particular, slit settings *slits_at_Tlo*, *Tlo*, *Thi*,
        and *slits_below*, and *slits_above* are used to define the
        angular divergence.

        .. Note::
             This function ignores any resolution information stored in
             the file, such as dQ, dT or dL columns, and instead uses the
             defined instrument parameters to calculate the resolution.

        """
        # Load the data
        data = np.loadtxt(filename).T
        if data.shape[0] == 2:
            Q, R = data
            dR = None
        elif data.shape[0] == 3:
            Q, R, dR = data
        elif data.shape[0] == 4:
            Q, dQ, R, dR = data
        elif data.shape[0] == 5:
            Q, dQ, R, dR, L = data
        if "Q" not in kw: kw["Q"] = Q
        T, dT, L, dL = self.resolution(**kw)
        kw.update(dict(T=T, dT=dT, L=L, dL=dL, data=(R, dR),
                       radiation=self.radiation))
        return make_probe(**kw)

class GenericPulsed(Pulsed):
    def load(self, filename, **kw):
        """
        Load the data, returning the associated probe.  This probe will
        contain Q, angle, wavelength, measured reflectivity and the
        associated uncertainties.

        You can override instrument parameters using key=value.
        In particular, slit settings *slits* and *T* define the
        angular divergence.
        """
        # Load the data
        data = np.loadtxt(filename).T
        Q, dQ, R, dR, L = data
        dL = binwidths(L)
        T = kw.pop('T', QL2T(Q, L))
        T, dT, L, dL = self.resolution(L=L, dL=dL, T=T, **kw)
        return make_probe(T=T, dT=dT, L=L, dL=dL, data=(R, dR),
                          radiation=self.radiation, **kw)
'''



"""
Basic interactor for Reflectometry profile.
"""

from .binder import pixel_to_data
from .config import active_color


def safecall(fn):
    def wrapped(self, *args, **kw):
        try:
            fn(self, *args, **kw)
        except:
            if self._debug:
                import sys, traceback

                traceback.print_exc()
                sys.exit(1)
            else:
                raise

    return wrapped


# ---------------------- Base interactors ----------------------
# GUI starts here
# Other interactor will inherit this
# --------------------------------------------------------------


class BaseInteractor(object):
    """
    Common base class for profile interactors

    Individual interactors need the following functions:

        save(ev)    - save the current state for later restore
        restore(ev) - restore the old state
        drag(ev)    - move the interactor to position ev.xdata,ev.ydata
        update_markers() - redraw the interactors

    If something special is required while dragging:

        drag_start(ev)  - drag started
        drag_cancel(ev) - drag cancelled with escape key
        drag_done(ev)   - drag completed

    The following are provided by the base class:

        connect_markers(markers) - register callbacks for all markers
        clear_markers() - remove all items in self.markers

    The actions associated with the interactor are:
        _onHilite(ev) - enter/leave event processing
        _onLeave(ev) - enter/leave event processing
        _onClick(ev) - mouse click: calls save()
        _onRelease(ev) - mouse click ends: calls moveend()
        _onDrag(ev) - mouse move: calls move() or restore()
        _onKey(ev) - keyboard move: calls move() or restore()

    Interactor attributes:

        profile  - the profile containing the interaactor
        color    - color of the interactor in non-active state
        markers  - list of handles for the interactor
    """

    _debug = False

    def __init__(self, profile):
        self.profile = profile
        self.markers = []

        # Remember if we are dragging during motion
        self._dragging = False

        # Remember the last location for keyboard dragging
        self._arrow_x, self._arrow_y = None, None

    def connect_markers(self, markers):
        """
        Connect markers to callbacks
        """
        connect = self.profile.connect
        for h in markers:
            connect("enter", h, self._onEnter)
            connect("leave", h, self._onLeave)
            connect("click", h, self._onClick)
            connect("release", h, self._onRelease)
            connect("drag", h, self._onDrag)
            connect("key", h, self._onKey)

    def clear_markers(self):
        """Clear old markers and interfaces."""
        connect = self.profile.connect
        for h in self.markers:
            h.remove()
        if self.markers:
            connect.clear(*self.markers)
        self.markers = []

    # ======================================
    def update_markers(self):
        """
        Configuration has changed: update marker positions
        """

    def save(self, ev):
        """
        Save the current state.

        Save is triggered by both a mouse click and a mouse release.
        """

    def restore(self, ev):
        """
        Restore the saved state
        """

    def drag(self, ev):
        """
        Move the interactor relative to the current state.

        Drag operations are usually triggered by the mouse, using the following
        call pattern::

            save()
            drag_start()
            drag()
            drag()
            ...
            drag_done()
            save()

        The final save allows keyboard actions to happen on the current
        interactor after the mouse manipulation has stopped.  Note that
        the drag actions are happening outside the drag_start/drag_done.
        Note also that the event itself will be marked as a keyboard
        event.

            save()
            drag_start() # mouse click
            drag_done()  # mouse release
            save()
            drag()       # arrow key
            drag()       # arrow key
            drag()       # arrow key
            restore()    # escape key

        When the escape key is pressed during drag, or the mouse leaves the
        screen, the drag may be cancelled::

            save()
            drag_start()
            drag()
            drag()
            ...
            drag_cancel()
            restore()
        """

    def drag_start(self, ev):
        """
        Start drag.

        Normally don't need to do anything here since save is triggered
        automatically.
        """

    def drag_cancel(self, ev):
        """
        Cancel drag.

        Normally don't need to do anything since restore is triggered
        automatically.
        """

    def drag_done(self, ev):
        """
        End drag.

        Normally don't need to do anything since the value was updated
        by drag.
        """

    # ======================================#
    # Event processors --- do not override #
    # ======================================#
    def _onEnter(self, event):
        """
        Hilite the artist reporting the event, indicating that it is
        ready to receive a click.
        """
        self._color = event.artist.get_color()
        event.artist.set_color(active_color)
        # TODO: use overlay on Hilite and remove it on Leave
        self.profile.draw_idle()

        return True

    def _onLeave(self, event):
        """
        Restore the artist to the original colour when the cursor leaves.
        """
        event.artist.set_color(self._color)
        self.profile.draw_idle()
        return True

    @safecall
    def _onClick(self, event):
        """
        Prepare to move the artist.  Calls save() to preserve the state for
        later restore().
        """
        # Make sure the x,y data use the coordinate system of the
        # artist rather than the default axes coordinates.
        transform = event.artist.get_transform()
        xy = pixel_to_data(transform, event.x, event.y)
        event.xdata, event.ydata = xy

        # save current position and event
        self.save(event)

        self.drag_start(event)
        self._dragging = True

        return True

    @safecall
    def _onRelease(self, event):
        """
        Release the mouse
        """
        # We are done the click-drag operation
        self.drag_done(event)
        self._dragging = False
        # self.profile.draw_idle()

        # Prepare for keyboard adjustment
        self._arrow_trans = event.artist.get_transform()
        self._arrow_x = event.xdata
        self._arrow_y = event.ydata

        return True

    @safecall
    def _onDrag(self, event):
        """
        Move the artist.  Calls move() to update the state, or restore() if
        the mouse leaves the window.
        """

        inside, _ = self.profile.axes.contains(event)

        if inside:
            # In case of no click and just drag
            if not self._dragging:
                print("We have a drag without a click")
                self._dragging = True
            self.drag(event)
        else:
            # Dragging left the canvas: undo the change
            self.drag_cancel(event)
            self.restore(event)

        # update model
        self.profile.update()

        return True

    @safecall
    def _onKey(self, event):
        """
        Respond to keyboard events.  Arrow keys move the widget.  Escape
        restores it to the position before the last click.

        Calls move() to update the state.  Calls restore() on escape.

        Returns True if key is handled
        """
        if event.key == "escape":
            if self._dragging:
                self.drag_cancel(event)
            self.restore(event)
            self.profile.update()
            return True

        elif event.key in ["up", "down", "right", "left"]:
            if self._dragging:
                return True

            if not (hasattr(self, "_arrow_x") and hasattr(self, "_arrow_y")):
                print("got key event without having location")
                return True

            step = 0.2 if event.control else 1
            x, y = self._arrow_x, self._arrow_y
            px, py = self._arrow_trans.transform_point((x, y))
            if event.key == "up":
                py += step
            elif event.key == "down":
                py -= step
            elif event.key == "right":
                px += step
            elif event.key == "left":
                px -= step
            x, y = self._arrow_trans.inverted().transform_point((px, py))
            self._arrow_x, self._arrow_y = x, y

            # update the state
            event.xdata, event.ydata = x, y
            # TODO: sound bell when at limits
            self.drag(event)
            self.profile.update()

            return True

        else:
            return False

        raise RuntimeError("Unreachable code")



import numba
from .clone_module import clone_module

MODULE = clone_module("refl1d.lib.python.convolve_sampled")

convolve_point_sampled = numba.njit(cache=True)(MODULE.convolve_point_sampled)
MODULE.convolve_point_sampled = convolve_point_sampled

convolve_sampled = numba.njit(cache=True)(MODULE.convolve_sampled)



"""
Reflectivity plugin for fitting GUI.

Note that the fitting infrastructure is still heavily tied to the reflectivity
modeling program, and this represents only the first tiny steps to separating
the two.
"""

__all__ = ["data_view", "model_view", "new_model", "calc_errors", "show_errors"]

import numpy as np

from . import names as refl
from . import __version__
from .errors import calc_errors, show_errors

# List of modules that contain dataclasses for the saved json file format
from . import names, material, model, magnetism, probe, experiment


# These are names used by the driver
def data_view():
    from .view.data_view import DataView

    return DataView


def model_view():
    from .view.model_view import ModelView

    return ModelView


def load_model(filename):
    # TODO: bumps plugin api needs to allow options for loader
    options = None
    if filename.endswith(".so") or filename.endswith(".dll") or filename.endswith(".dyld"):
        from . import garefl

        problem = garefl.load(filename)
        return problem
    elif filename.endswith(".staj") or filename.endswith(".sta"):
        from .stajconvert import load_mlayer

        return refl.FitProblem(load_mlayer(filename))
        # fit_all(problem.fitness, pmp=20)
    elif filename.endswith(".zip"):
        from bumps.fitproblem import load_problem

        # Note: bumps.vfs.vfs_init() must be called very early
        try:
            from bumps.vfs import ZipFS
        except ImportError:
            raise NotImplementedError("need newer bumps to load model from zip")
        with ZipFS(filename) as zf:
            for f in zf.filelist:
                if f.filename.endswith(".py"):
                    return load_problem(f.filename, options=options)
    elif filename.endswith(".json"):
        from bumps.serialize import load

        return load(filename)
    else:
        return None


def save_json(problem, basename):
    from bumps.serialize import save

    json_filename = basename + "-expt.json"
    save(json_filename, problem)


def new_model():
    stack = refl.silicon(0, 10) | refl.air
    instrument = refl.NCNR.NG1()
    probe = instrument.probe(T=np.linspace(0, 5, 200), Tlo=0.2, slits_at_Tlo=2)
    M = refl.Experiment(sample=stack, probe=probe)
    problem = refl.FitProblem(M)
    return problem



"""
Common materials in reflectometry experiments along with densities.

By name::

    air, water, heavywater, lightheavywater, silicon, sapphire, gold
    permalloy

By formula::

    H2O, D2O, DHO, Si, Al2O3, Au, Ni8Fe2

If you want to adjust the density you will need to make your own copy of
these materials.  For example, for permalloy::

    >>> NiFe=Material(permalloy.formula, density=permalloy.bulk_density)
    >>> NiFe.density.pmp(10)  # Let density vary by 10% from bulk value
    Parameter(permalloy density)
"""

import periodictable
from .material import Vacuum, Material

__all__ = [
    "air",
    "water",
    "H2O",
    "heavywater",
    "D2O",
    "lightheavywater",
    "DHO",
    "silicon",
    "Si",
    "sapphire",
    "Al2O3",
    "gold",
    "Au",
    "permalloy",
    "Ni8Fe2",
]

# Set the bulk density of heavy water assuming it has the same packing
# fraction as water, but with heavier H[2] substituted for H[1].
rho_D2O = periodictable.formula("D2O").mass / periodictable.formula("H2O").mass
rho_DHO = periodictable.formula("DHO").mass / periodictable.formula("H2O").mass

air = Vacuum()
water = H2O = Material("H2O", density=1, name="water")
heavywater = D2O = Material("D2O", density=rho_D2O)
lightheavywater = DHO = Material("DHO", density=rho_DHO)
silicon = Si = Material("Si")
sapphire = Al2O3 = Material("Al2O3", density=3.965, name="sapphire")
gold = Au = Material("Au", name="gold")
permalloy = Ni8Fe2 = Material("Ni8Fe2", density=8.692, name="permalloy")



import numba
import importlib.util


def clone_module(original_module: str):
    spec = importlib.util.find_spec(original_module)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module



# Below taken from the GUI profile plot methods in Refl1d - modified to be stand alone
# This is to allow for the layer labels to be easily added to the SLD plots

# Originally written for use in a jupyter notebook
# Adapted to act as a prototype for the profile plotting in the new GUI for refl1d

from typing import List, TypedDict, TYPE_CHECKING
from refl1d.experiment import Experiment
from numpy import inf
import numpy as np

if TYPE_CHECKING:
    import plotly.graph_objs as go

from .colors import COLORS as MULTI_PLOT_COLORS

SINGLE_PLOT_COLORS = ("black", "blue", "green", "gold")

SINGLE_PLOT_STYLES = ["solid"] * 4
MULTI_PLOT_STYLES = ["solid", "dash", "dashdot", "dot"]

# === Sample information ===


class FindLayers:
    def __init__(self, experiment, axes=None, x_offset=0):
        self.experiment = experiment
        self.axes = axes
        self.x_offset = x_offset
        self._show_labels = True
        self._show_boundaries = True

        self._find_layer_boundaries()

    def find(self, z):
        return self.experiment.sample.find(z - self.x_offset)

    def _find_layer_boundaries(self):
        offset = self.x_offset
        boundary = [-inf, offset]
        if hasattr(self.experiment, "sample"):
            for L in self.experiment.sample[1:-1]:
                dx = L.thickness.value
                offset += dx
                boundary.append(offset)
        boundary.append(inf)
        self.boundary = np.asarray(boundary)

    def sample_labels(self):
        return [L.name for L in self.experiment.sample]

    def sample_layer(self, n):
        return self.experiment.sample[n]

    def label_offsets(self):
        z = self.boundary[1:-1]
        left = -20
        middle = [(a + b) / 2.0 for a, b in zip(z[:-1], z[1:])]
        right = z[-1] + 20
        return [left] + middle + [right]

    def reset_markers_plotly(self):
        """
        Reset all markers, for plotly plots
        """
        import plotly.graph_objs as go

        # self.clear_markers()
        if not isinstance(self.axes, go.Figure):
            raise ValueError("reset_markers_plotly can only be used with type: axes=plotly.graph_objs.Figure")

        fig = self.axes

        # Add bars
        fittable = [self.experiment.sample[idx].thickness.fittable for idx, _ in enumerate(self.boundary[1:-1])]
        fittable[0] = False  # First interface is not fittable

        for f, z_pos in zip(fittable, self.boundary[1:-1]):
            if not f:
                line_dict = dict(dash="dash")
            else:
                line_dict = dict(dash="solid")

            fig.add_vline(
                x=z_pos,
                opacity=0.75,
                line=line_dict,
            )

        # for f,m in zip(fittable,self.markers):
        #     if not f: m.set(linestyle='--', linewidth=1.25)
        # self.connect_markers(m for f,m in zip(fittable,self.markers) if f)

        # Add labels
        offsets = self.label_offsets()
        labels = self.sample_labels()
        for label_pos, label in zip(offsets, labels):
            fig.add_annotation(
                text=label, textangle=-30, x=label_pos, yref="paper", y=1.0, yanchor="bottom", showarrow=False
            )


def generate_best_profile(model: Experiment):
    if model.ismagnetic:
        best = model.magnetic_smooth_profile(dz=0.5)
    else:
        best = model.smooth_profile()
    return best


# ============================================================================= #
# Plotting script below
# ============================================================================= #


class ModelSpec(TypedDict):
    model_index: int
    sample_index: int


class PlotItem(TypedDict):
    model: Experiment
    spec: ModelSpec
    color_index: int


def plot_multiple_sld_profiles(plot_items: List[PlotItem]):
    import plotly.graph_objs as go

    fig = go.Figure()
    nplots = len(plot_items)
    multiplot = nplots > 1
    for plot_item in plot_items:
        model = plot_item["model"]
        spec = plot_item["spec"]
        color_index = plot_item["color_index"]
        model_index = spec["model_index"]
        sample_index = spec["sample_index"]
        if multiplot:
            colors = [MULTI_PLOT_COLORS[color_index % len(MULTI_PLOT_COLORS)]] * 4
        else:
            colors = SINGLE_PLOT_COLORS
        line_styles = MULTI_PLOT_STYLES if multiplot else SINGLE_PLOT_STYLES
        legendgroup = f"{model_index}:{sample_index} ({model.name})" if multiplot else None
        plot_sld_profile_plotly(model, fig, colors=colors, line_styles=line_styles, legendgroup=legendgroup)
    return fig


def plot_sld_profile_plotly(model, fig, colors=SINGLE_PLOT_COLORS, line_styles=SINGLE_PLOT_STYLES, legendgroup=None):
    if model.ismagnetic:
        z_best, rho_best, irho_best, rhoM_best, thetaM_best = generate_best_profile(model)
        yaxis_title = "SLD: ρ, ρ<sub>i</sub>, ρ<sub>M</sub> / 10<sup>-6</sup> Å<sup>-2</sup>"
    else:
        z_best, rho_best, irho_best = generate_best_profile(model)
        yaxis_title = "SLD: ρ, ρ<sub>i</sub> / 10<sup>-6</sup> Å<sup>-2</sup>"

    hovertemplate = (
        f"(%{{x}}, %{{y}})<br>{legendgroup if legendgroup is not None else ''} %{{data.hovertext}}<extra></extra>"
    )

    fig.add_scatter(
        x=z_best,
        y=rho_best,
        name="ρ",
        legendgroup=legendgroup,
        legendgrouptitle_text=legendgroup,
        hovertemplate=hovertemplate,
        hovertext="SLD",
        line={
            "color": colors[0],
            "dash": line_styles[0],
        },
    )

    fig.add_scatter(
        x=z_best,
        y=irho_best,
        name="ρ<sub>i</sub>",
        legendgroup=legendgroup,
        hovertemplate=hovertemplate,
        hovertext="Im. SLD",
        line={
            "color": colors[1],
            "dash": line_styles[1],
        },
    )

    if model.ismagnetic:
        fig.add_scatter(
            x=z_best,
            y=rhoM_best,
            name="ρ<sub>M</sub>",
            legendgroup=legendgroup,
            hovertemplate=hovertemplate,
            hovertext="Mag. SLD",
            line={
                "color": colors[2],
                "dash": line_styles[2],
            },
        )

        fig.add_scatter(
            x=z_best,
            y=thetaM_best,
            name="θ<sub>M</sub>",
            yaxis="y2",
            legendgroup=legendgroup,
            hovertemplate=hovertemplate,
            hovertext="Theta M",
            line={
                "color": colors[3],
                "dash": line_styles[3],
            },
        )

        # TODO: need to make axis scaling for thetaM dependent on if thetaM exceeds 0-360
        fig.update_layout(
            yaxis2={
                "title": {"text": "Magnetic Angle θ<sub>M</sub> / °"},
                "type": "linear",
                "autorange": False,
                "range": [0, 360],
                "anchor": "x",
                "overlaying": "y",
                "side": "right",
                "ticks": "inside",
                # 'ticklen': 20,
            }
        )

    fig.update_layout(uirevision=1, plot_bgcolor="white")
    fig.update_layout(
        xaxis={
            "title": {"text": "depth (Å)"},
            "type": "linear",
            "autorange": True,
            # 'gridcolor': "Grey",
            "ticks": "inside",
            # 'ticklen': 20,
            "showline": True,
            "linewidth": 1,
            "linecolor": "black",
            "mirror": "ticks",
            "side": "bottom",
        }
    )

    fig.update_layout(
        yaxis={
            "title": {"text": yaxis_title},
            "exponentformat": "e",
            "showexponent": "all",
            "type": "linear",
            "autorange": True,
            # 'gridcolor': "Grey",
            "ticks": "inside",
            # 'ticklen': 20,
            "showline": True,
            "linewidth": 1,
            "linecolor": "black",
            "mirror": True,
        }
    )

    fig.update_layout(margin={"l": 75, "r": 50, "t": 50, "b": 75, "pad": 4})

    fig.update_layout(
        legend={
            "x": -0.1,
            "bgcolor": "rgba(255,215,0,0.15)",
            # "traceorder": "reversed"
        }
    )

    marker_positions = FindLayers(model, axes=fig)
    marker_positions.reset_markers_plotly()

    # fig.show(renderer='firefox')
    return fig


# def plot_contours(model, title, savepath=None, nsamples=1000, show_contours=True, show_mag=True, savetitle=None,
#                   store=None, ultranest=False, dream=False, align=0):
#     if show_contours:
#         # data, columns = generate_contour_data(model)
#         if ultranest:
#             points_array = model.post_samples
#             sub_array = points_array[np.sort(np.random.randint(points_array.shape[0], size=nsamples)), :]
#             errors = calc_errors(model.problem, sub_array)
#             data, columns = generate_profile_data(errors)
#         # if this comment is removed the above code will not overwrite the previous reference to the errors object
#         # and it will act as a bad memory leak - using more ram everytime the function is called
#         if dream:
#             state, points = load_dream_state_notebook(problem=model.problem, store=store)
#             errors = calc_errors(model.problem, points[-nsamples:-1])
#             data, columns = generate_profile_data(errors, align)
#
#     # get best profile manually as we are not using DREAM state
#     if show_mag:
#         z_best, rho_best, irho_best, rhoM_best, thetaM_best = generate_best_profile(model)
#     else:
#         z_best, rho_best, irho_best = generate_best_profile(model)
#
#     fig, ax = plt.subplots(1, 1, figsize=(8, 6))
#
#     # rho
#     if show_contours:
#         z, best, sig2lo, sig2hi, siglo, sighi = data[0]
#         ax.plot(z, best, label=r"$\rho$ SLD", color="orange")
#         ax.fill_between(z, siglo, sighi, alpha=0.5, color="orange", label=r"$\rho$ SLD $1\sigma$")
#         ax.fill_between(z, sig2lo, sig2hi, alpha=0.25, color="orange", label=r"$\rho$ SLD $2\sigma$")
#     else:
#         ax.plot(z_best, rho_best, label=r"$\rho$ SLD", color="orange")
#
#     # irho
#     # ax.plot(z_best, irho_best, label=r"$\rho_{i}$ Im SLD")
#     # if show_contours:
#     #     z, best, sig2lo, sig2hi, siglo, sighi = data[1]
#     #     ax.fill_between(z, siglo, sighi, alpha=0.5)
#     #     ax.fill_between(z, sig2lo, sig2hi, alpha=0.25)
#     if show_mag:
#         if show_contours:
#             z, best, sig2lo, sig2hi, siglo, sighi = data[2]
#             ax.plot(z, best, label=r"$\rho_{M}$ mSLD", color="b")
#             ax.fill_between(z, siglo, sighi, alpha=0.5, color="b", label=r"$\rho_{M}$ mSLD $1\sigma$")
#             ax.fill_between(z, sig2lo, sig2hi, alpha=0.25, color="b", label=r"$\rho_{M}$ mSLD $2\sigma$")
#         else:
#             ax.plot(z_best, rhoM_best, label=r"$\rho_{M}$ mSLD", color="b")
#
#     # ax.legend(fontsize=14, loc='upper right', framealpha=0.5, ncol=2)
#     ax.legend(fontsize=16, framealpha=0.5, ncol=3)
#
#     ax.grid(True)
#     ax.set_xlabel(r"z $\left(\AA\right)$", fontsize=18)
#     ax.set_ylabel(r"$\rho, \rho_{i} \left(10^{-6}\AA^{-2}\right)$", fontsize=18)
#     if show_mag:
#         ax.set_ylabel(r"$\rho, \rho_{M} \left(10^{-6}\AA^{-2}\right)$", fontsize=18)
#     ax.tick_params(axis='y', labelsize=16)
#     ax.tick_params(axis='x', labelsize=16)
#
#     experiment = model.determine_problem_fitness()
#
#     layer_markers = FindLayers(experiment, axes=ax)
#     layer_markers.reset_markers()
#     fig.suptitle(title)
#     # if np.max(rho_best) > 75:
#     #     top = np.max(rho_best)*1.15
#     # else:
#     #     top =75
#     # ax.set_ylim(top=top)
#     if savepath:
#         if savetitle:
#             title = savetitle
#         plt.savefig(f"{savepath}/{title}.png")



"""
Basic reflectometry calculations

Slab model reflectivity calculator with optional absorption and roughness.
The function reflectivity_amplitude returns the complex waveform.
Slab model with supporting magnetic scattering.  The function
magnetic_reflectivity returns the complex reflection for the four
spin polarization cross sections [++, +-, -+, --].  The function
unpolarized_magnetic returns the expected magnitude for a measurement
of the magnetic scattering using an unpolarized beam.
"""

from __future__ import print_function

from functools import reduce

# __doc__ = "Fundamental reflectivity calculations"
__author__ = "Paul Kienzle"
__all__ = [
    "reflectivity",
    "reflectivity_amplitude",
    "magnetic_reflectivity",
    "magnetic_amplitude",
    "unpolarized_magnetic",
    "convolve",
]

import numpy as np
from numpy import pi, sin, cos, conj, radians, sqrt, exp, fabs

BASE_GUIDE_ANGLE = 270.0


def _dense(x, dtype="d"):
    return np.ascontiguousarray(x, dtype)


def reflectivity(*args, **kw):
    """
    Calculate reflectivity $|r(k_z)|^2$ from slab model.

    :Parameters :
        *depth* : float[N] | |Ang|
            Thickness of the individual layers (incident and substrate
            depths are ignored)
        *sigma* : float OR float[N-1] | |Ang|
            Interface roughness between the current layer and the next.
            The final layer is ignored.  This may be a scalar for fixed
            roughness on every layer, or None if there is no roughness.
        *rho*, *irho* : float[N] OR float[N, K] | |1e-6/Ang^2|
            Real and imaginary scattering length density.  Use multiple
            columns when you have kz-dependent scattering length densities,
            and set rho_offset to select the appropriate one.  Data should
            be stored in column order.
        *kz* : float[M] | |1/Ang|
            Points at which to evaluate the reflectivity
        *rho_index* : integer[M]
            *rho* and *irho* columns to use for the various kz.

    :Returns:
        *R* | float[M]
            Reflectivity magnitude.

    This function does not compute any instrument resolution corrections.
    """
    r = reflectivity_amplitude(*args, **kw)
    return (r * conj(r)).real


def reflectivity_amplitude(
    kz=None,
    depth=None,
    rho=None,
    irho=0,
    sigma=0,
    rho_index=None,
):
    r"""
    Calculate reflectivity amplitude $r(k_z)$ from slab model.

    :Parameters :
        *depth* : float[N] | |Ang|
            Thickness of the individual layers (incident and substrate
            depths are ignored)
        *sigma* = 0 : float OR float[N-1] | |Ang|
            Interface roughness between the current layer and the next.
            The final layer is ignored.  This may be a scalar for fixed
            roughness on every layer, or None if there is no roughness.
        *rho*, *irho* = 0: float[N] OR float[N, K] | |1e-6/Ang^2|
            Real and imaginary scattering length density.  Use multiple
            columns when you have kz-dependent scattering length densities,
            and set *rho_index* to select amongst them.  Data should be
            stored in column order.
        *kz* : float[M] | |1/Ang|
            Points at which to evaluate the reflectivity
        *rho_index* = 0 : integer[M]
            *rho* and *irho* columns to use for the various kz.

    :Returns:
        *r* | complex[M]
            Complex reflectivity waveform.

    This function does not compute any instrument resolution corrections.
    """
    from .refllib import backend

    kz = _dense(kz, "d")
    if rho_index is None:
        rho_index = np.zeros(kz.shape, "i")
    else:
        rho_index = _dense(rho_index, "i")

    depth = _dense(depth, "d")
    if np.isscalar(sigma):
        sigma = sigma * np.ones(len(depth) - 1, "d")
    else:
        sigma = _dense(sigma, "d")
    rho = _dense(rho, "d")
    # promote rho and irho to 2d, for multi-wavelength
    if rho.ndim == 1:
        rho.resize((1, rho.shape[0]))
    if np.isscalar(irho):
        irho = irho * np.ones_like(rho)
    else:
        irho = _dense(irho, "d")
    if irho.ndim == 1:
        irho.resize((1, irho.shape[0]))

    # print(irho.shape, irho[:,0], irho[:,-1])
    irho = abs(irho) + 1e-30
    # irho[irho < 0] = 0.
    # print depth.shape, rho.shape, irho.shape, sigma.shape
    # print depth.dtype, rho.dtype, irho.dtype, sigma.dtype
    r = np.empty(kz.shape, "D")
    # print "amplitude", depth, rho, kz, rho_index
    # print depth.shape, sigma.shape, rho.shape, irho.shape, kz.shape
    backend.reflectivity_amplitude(depth, sigma, rho, irho, kz, rho_index, r)

    return r


def magnetic_reflectivity(*args, **kw):
    """
    Magnetic reflectivity for slab models.

    Returns the expected values for the four polarization cross
    sections (++, +-, -+, --).
    Return reflectivity R^2 from slab model with sharp interfaces.
    returns reflectivities.

    The parameters are as follows:

    kz (|1/Ang|)
        points at which to evaluate the reflectivity
    depth (|Ang|)
        thickness of the individual layers (incident and substrate
        depths are ignored)
    rho (|1e-6/Ang^2|)
        Scattering length density.
    irho (|1e-6/Ang^2|)
        absorption. Defaults to 0.
    rho_m (microNb)
        Magnetic scattering length density correction.
    theta_m (degrees)
        Angle of the magnetism within the layer.
    sigma (|Ang|)
        Interface roughness between the current layer and the next.
        The final layer is ignored.  This may be a scalar for fixed
        roughness on every layer, or None if there is no roughness.
    wavelength (|Ang|)
        Incident wavelength (only affects absorption).  May be a vector.
        Defaults to 1.
    Aguide (degrees)
        Angle of the guide field; -90 is the usual case

    This function does not compute any instrument resolution corrections.
    Interface diffusion, if present, uses the Nevot-Croce approximation.

    Use magnetic_amplitude to return the complex waveform.
    """
    r = magnetic_amplitude(*args, **kw)
    return [(z * z.conj()).real for z in r]


def unpolarized_magnetic(*args, **kw):
    """
    Returns the average of magnetic reflectivity for all cross-sections.

    See :class:`magnetic_reflectivity <refl1d.reflectivity.magnetic_reflectivity>` for details.
    """
    return reduce(np.add, magnetic_reflectivity(*args, **kw)) / 2.0


B2SLD = 2.31604654  # Scattering factor for B field 1e-6/


def magnetic_amplitude(
    kz,
    depth,
    rho,
    irho=0,
    rhoM=0,
    thetaM=0,
    sigma=0,
    Aguide=-90,
    H=0,
    rho_index=None,
):
    """
    Returns the complex magnetic reflectivity waveform.

    See :class:`magnetic_reflectivity <refl1d.reflectivity.magnetic_reflectivity>` for details.
    """
    from .refllib import backend

    kz = _dense(kz, "d")
    if rho_index is None:
        rho_index = np.zeros(kz.shape, "i")
    else:
        rho_index = _dense(rho_index, "i")
    n = len(depth)
    if np.isscalar(irho):
        irho = irho * np.ones(n, "d")
    if np.isscalar(rhoM):
        rhoM = rhoM * np.ones(n, "d")
    if np.isscalar(thetaM):
        thetaM = thetaM * np.ones(n, "d")
    if np.isscalar(sigma):
        sigma = sigma * np.ones(n - 1, "d")

    # kz = -kz
    # depth, rho, irho, sigma, rhoM, thetaM = [v[::-1] for v in (depth, rho, irho, sigma, rhoM, thetaM)]
    depth, rho, irho, sigma = [_dense(a, "d") for a in (depth, rho, irho, sigma)]
    # np.set_printoptions(linewidth=1000)
    # print(np.vstack((depth, np.hstack((sigma, np.nan)), rho, irho, rhoM, thetaM)).T)

    sld_b, u1, u3 = calculate_u1_u3(H, rhoM, thetaM, Aguide)

    R1, R2, R3, R4 = [np.empty(kz.shape, "D") for pol in (1, 2, 3, 4)]
    backend.magnetic_amplitude(depth, sigma, rho, irho, sld_b, u1, u3, kz, rho_index, R1, R2, R3, R4)
    return R1, R2, R3, R4


def calculate_u1_u3(H, rhoM, thetaM, Aguide):
    from .refllib import backend

    rhoM, thetaM = (_dense(rhoM, "d")).copy(), _dense(np.radians(thetaM), "d")
    n = len(rhoM)
    u1, u3 = np.empty(n, "D"), np.empty(n, "D")
    backend.calculate_u1_u3(H, rhoM, thetaM, Aguide, u1, u3)

    return rhoM, u1, u3


def calculate_u1_u3_py(H, rhoM, thetaM, Aguide):
    rotate_M = True

    EPS = np.finfo("f").tiny  # not 1e-20 # epsilon offset for divisions.
    thetaM = radians(thetaM)
    phiH = radians(Aguide - BASE_GUIDE_ANGLE)
    thetaH = np.pi / 2.0  # by convention, H is in y-z plane so theta = pi/2

    sld_h = B2SLD * H
    sld_m_x = rhoM * np.cos(thetaM)
    sld_m_y = rhoM * np.sin(thetaM)
    sld_m_z = 0.0  # by Maxwell's equations, H_demag = mz so we'll just cancel it here
    # The purpose of AGUIDE is to rotate the z-axis of the sample coordinate
    # system so that it is aligned with the quantization axis z, defined to be
    # the direction of the magnetic field outside the sample.
    if rotate_M:
        # rotate the M vector instead of the transfer matrix!
        # First, rotate the M vector about the x axis:
        new_my = sld_m_z * sin(radians(Aguide)) + sld_m_y * cos(radians(Aguide))
        new_mz = sld_m_z * cos(radians(Aguide)) - sld_m_y * sin(radians(Aguide))
        sld_m_y, sld_m_z = new_my, new_mz
        sld_h_x = sld_h_y = 0.0
        sld_h_z = sld_h
        # Then, don't rotate the transfer matrix
        Aguide = 0.0
    else:
        sld_h_x = sld_h * np.cos(thetaH)  # zero
        sld_h_y = sld_h * np.sin(thetaH) * np.cos(phiH)
        sld_h_z = sld_h * np.sin(thetaH) * np.sin(phiH)

    sld_b_x = sld_h_x + sld_m_x
    sld_b_y = sld_h_y + sld_m_y
    sld_b_z = sld_h_z + sld_m_z

    # avoid divide-by-zero:
    sld_b_x += EPS * (sld_b_x == 0)
    sld_b_y += EPS * (sld_b_y == 0)

    # add epsilon to y, to avoid divide by zero errors?
    sld_b = np.sqrt(sld_b_x**2 + sld_b_y**2 + sld_b_z**2)
    u1_num = +sld_b + sld_b_x + 1j * sld_b_y - sld_b_z
    u1_den = +sld_b + sld_b_x - 1j * sld_b_y + sld_b_z
    u3_num = -sld_b + sld_b_x + 1j * sld_b_y - sld_b_z
    u3_den = -sld_b + sld_b_x - 1j * sld_b_y + sld_b_z

    u1 = u1_num / u1_den
    u3 = u3_num / u3_den
    # print "u1", u1
    # print "u3", u3
    return sld_b, u1, u3


def convolve(xi, yi, x, dx, resolution="normal"):
    r"""
    Apply x-dependent gaussian resolution to the theory.

    Returns convolution y[k] of width dx[k] at points x[k].

    The theory function is a piece-wise linear spline which does not need to
    be uniformly sampled.  The theory calculation points *xi* should be dense
    enough to capture the "wiggle" in the theory function, and should extend
    beyond the ends of the data measurement points *x*. Convolution at the
    tails is truncated and normalized to area of overlap between the resolution
    function in case the theory does not extend far enough.

    *resolution* is 'normal' (default) or 'uniform'. Note that the uniform
    distribution uses the $1-\sigma$ equivalent distribution width which is
    $1/\sqrt{3}$ times the width of the rectangle.
    """
    from .refllib import backend

    xi, yi, x, dx = _dense(xi), _dense(yi), _dense(x), _dense(dx)
    y = np.empty_like(x)
    if resolution == "uniform":
        backend.convolve_uniform(xi, yi, x, dx, y)
    else:
        backend.convolve_gaussian(xi, yi, x, dx, y)
    return y


def convolve_sampled(xi, yi, xp, yp, x, dx):
    """
    Apply x-dependent arbitrary resolution function to the theory.

    Returns convolution y[k] of width dx[k] at points x[k].

    Like :func:`convolve`, the theory *(xi, yi)* is represented as a
    piece-wise linear spline which should extend beyond the data
    measurement points *x*.  Instead of a gaussian resolution function,
    resolution *(xp, yp)* is also represented as a piece-wise linear
    spline.
    """
    from .refllib import backend

    x = _dense(x)
    y = np.empty_like(x)
    backend.convolve_sampled(_dense(xi), _dense(yi), _dense(xp), _dense(yp), x, _dense(dx), y)
    return y


def test_uniform():
    xi = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    yi = [1, 3, 1, 2, 7, 3, 1, 2, 1, 3]
    _check_uniform("uniform aligned", xi, yi, x=[2, 4, 6, 8], dx=1)
    _check_uniform("uniform unaligned", xi, yi, x=[2.5, 4.5, 6.5, 8.5], dx=1)
    _check_uniform("uniform wide", xi, yi, x=[2.5, 4.5, 6.5, 8.5], dx=3)
    # Check bad values
    ystar = convolve(xi, yi, x=[-3, 13], dx=[1 / np.sqrt(3)] * 2, resolution="uniform")
    assert (ystar == 0.0).all()

    xi = [1.1, 2.3, 2.8, 4.2, 5, 6, 7, 8, 9, 10]
    yi = [1, 3, 1, 2, 7, 3, 1, 2, 1, 3]
    _check_uniform("uniform unaligned", xi, yi, x=[2.5, 4.5, 6.5, 8.5], dx=1)


def _check_uniform(name, xi, yi, x, dx):
    # Note: using fixed dx since that's all _check_spline supports.
    ystar = convolve(xi, yi, x, [dx / np.sqrt(3)] * len(x), resolution="uniform")
    # print("xi", xi)
    # print("yi", yi)
    # print("x", x, "dx", dx)
    # print("ystar", ystar)
    xp = [-dx, -dx, dx, dx]
    yp = [0, 0.5 / dx, 0.5 / dx, 0]
    _check_spline(name, xi, yi, xp, yp, x, ystar)


def test_convolve_sampled():
    xi = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    yi = [1, 3, 1, 2, 1, 3, 1, 2, 1, 3]
    xp = [-1, 0, 1, 2, 3]
    yp = [1, 4, 3, 2, 1]
    _check_sampled("sampled aligned", xi, yi, xp, yp, dx=1)
    _check_sampled("sampled unaligned", xi, yi, _dense(xp) - 0.2000003, yp, dx=1)
    _check_sampled("sampled wide", xi, yi, xp, yp, dx=2)
    _check_sampled("sampled super wide", xi, yi, xp, yp, dx=10)


def _check_sampled(name, xi, yi, xp, yp, dx):
    ystar = convolve_sampled(xi, yi, xp, yp, xi, dx=np.full_like(xi, dx))
    xp = np.array(xp) * dx
    _check_spline(name, xi, yi, xp, yp, xi, ystar)


def _check_spline(name, xi, yi, xp, yp, x, ystar):
    step = 0.0001
    xpfine = np.arange(xp[0], xp[-1] + step / 10, step)
    ypfine = np.interp(xpfine, xp, yp)
    # make sure xfine is wide enough by adding an extra interval at the ends
    xfine = np.arange(xi[0] + xpfine[0] - 2 * step, xi[-1] + xpfine[-1] + 2 * step, step)
    yfine = np.interp(xfine, xi, yi, left=0, right=0)
    pidx = np.searchsorted(xfine, np.array(x) + xp[0])
    left, right = np.searchsorted(xfine, [xi[0], xi[-1]])

    conv = []
    for pk in pidx:
        norm_start = max(0, left - pk)
        norm_end = min(len(xpfine), right - pk)
        norm = step * np.sum(ypfine[norm_start:norm_end])
        conv.append(step * np.sum(ypfine * yfine[pk : pk + len(xpfine)]) / norm)

    # print("checking convolution %s"%(name, ))
    # print(" ".join("%7.4f"%yi for yi in ystar))
    # print(" ".join("%7.4f"%yi for yi in conv))
    assert all(abs(yi - fi) < 0.0005 for (yi, fi) in zip(ystar, conv)), name
